@@ -19,6 +19,7 @@ import json
 import logging
 import os
 import pprint
+import re
 import shutil
 import signal
 import socket
@@ -53,7 +54,6 @@ except AttributeError:
 
 
 websocket.logger = logging.getLogger("websocket")
-
 log = logging.getLogger("jujuclient")
 
 
@@ -183,7 +183,7 @@ class Connector(object):
             try:
                 return Connector.connect_socket(endpoint, cert_path)
             except socket.error as err:
-                if not err.errno in self.retry_conn_errors:
+                if err.errno not in self.retry_conn_errors:
                     raise
                 time.sleep(1)
                 continue
@@ -207,20 +207,61 @@ class Connector(object):
         return s
 
     def parse_env(self, env_name):
-        import yaml
         jhome = os.path.expanduser(
             os.environ.get('JUJU_HOME', '~/.juju'))
+
+        # Look in the cache file first.
+        cache_file = os.path.join(jhome, 'environments', 'cache.yaml')
         jenv = os.path.join(jhome, 'environments', '%s.jenv' % env_name)
+
+        if os.path.exists(cache_file):
+            try:
+                return jhome, self.environment_from_cache(env_name, cache_file)
+            except EnvironmentNotBootstrapped:
+                pass
+                # Fall through to getting the info from the jenv
         if not os.path.exists(jenv):
             raise EnvironmentNotBootstrapped(env_name)
+        return jhome, self.environment_from_jenv(jenv)
 
+    def environment_from_cache(self, env_name, cache_filename):
+        import yaml
+        with open(cache_filename) as fh:
+            data = yaml.safe_load(fh.read())
+            try:
+                # environment holds:
+                #   user, env-uuid, server-uuid
+                environment = data['environment'][env_name]
+                server = data['server-data'][environment['server-uuid']]
+                return {
+                    'user': environment['user'],
+                    'password': server['identities'][environment['user']],
+                    'environ-uuid': environment['env-uuid'],
+                    'server-uuid': environment['server-uuid'],
+                    'state-servers': server['api-endpoints'],
+                    'ca-cert': server['ca-cert'],
+                }
+            except KeyError:
+                raise EnvironmentNotBootstrapped(env_name)
+
+    def environment_from_jenv(self, jenv):
+        import yaml
         with open(jenv) as fh:
             data = yaml.safe_load(fh.read())
-            return jhome, data
+            return data
+
+    @staticmethod
+    def split_host_port(server):
+        m = re.match('(.*):(.*)', server)
+        if not m:
+            raise ValueError("Not an ipaddr/port {!r}".format(server))
+        address = m.group(1).strip("[]")
+        port = m.group(2)
+        return address, port
 
     def is_server_available(self, server):
         """ Given address/port, return true/false if it's up """
-        address, port = server.split(":")
+        address, port = Connector.split_host_port(server)
         try:
             socket.create_connection((address, port), 3)
             return True
@@ -253,20 +294,45 @@ class LogIterator(object):
 
 
 class RPC(object):
-
+    _upgrade_retry_delay_secs = 1
+    _upgrade_retry_count = 60
     _auth = False
     _request_id = 0
     _debug = False
     _reconnect_params = None
+
     conn = None
 
     def _rpc(self, op):
         if not self._auth and not op.get("Request") == "Login":
             raise LoginRequired()
-        if not 'Params' in op:
+        if 'Params' not in op:
             op['Params'] = {}
         op['RequestId'] = self._request_id
         self._request_id += 1
+        result = self._rpc_retry_if_upgrading(op)
+        if 'Error' in result:
+            # The backend disconnects us on err, bug: http://pad.lv/1160971
+            self.conn.connected = False
+            raise EnvError(result)
+        return result['Response']
+
+    def _rpc_retry_if_upgrading(self, op):
+        """If Juju is upgrading when the specified rpc call is made,
+        retry the call."""
+        retry_count = 0
+        result = {'Response': ''}
+        while retry_count <= self._upgrade_retry_count:
+            result = self._send_request(op)
+            if 'Error' in result and 'upgrade in progress' in result['Error']:
+                log.info("Juju upgrade in progress...")
+                retry_count += 1
+                time.sleep(self._upgrade_retry_delay_secs)
+                continue
+            break
+        return result
+
+    def _send_request(self, op):
         if self._debug:
             log.debug("rpc request:\n%s" % (json.dumps(op, indent=2)))
         self.conn.send(json.dumps(op))
@@ -274,12 +340,7 @@ class RPC(object):
         result = json.loads(raw)
         if self._debug:
             log.debug("rpc response:\n%s" % (json.dumps(result, indent=2)))
-
-        if 'Error' in result:
-            # The backend disconnects us on err, bug: http://pad.lv/1160971
-            self.conn.connected = False
-            raise EnvError(result)
-        return result['Response']
+        return result
 
     def login(self, password, user="user-admin"):
         """Login gets shared to watchers for reconnect."""
@@ -394,13 +455,23 @@ class Watcher(RPC):
 class TimeoutWatcher(Watcher):
     # A simple non concurrent watch using signals..
 
-    _timeout = None
+    def __init__(self, *args, **kw):
+        super(TimeoutWatcher, self).__init__(*args, **kw)
+        self.start_time = time.time()
+        self._timeout = 0
+
+    def time_remaining(self):
+        """Return number of seconds until this watch times out.
+
+        """
+        return int(self._timeout - (time.time() - self.start_time))
 
     def set_timeout(self, timeout):
+        self.start_time = time.time()
         self._timeout = timeout
 
     def next(self):
-        with self._set_alarm(self._timeout):
+        with self._set_alarm(self.time_remaining()):
             return super(TimeoutWatcher, self).next()
 
     # py3 compat
@@ -409,6 +480,9 @@ class TimeoutWatcher(Watcher):
     @classmethod
     @contextmanager
     def _set_alarm(cls, timeout):
+        if timeout < 0:
+            raise TimeoutError()
+
         try:
             handler = signal.getsignal(signal.SIGALRM)
             if callable(handler):
@@ -475,18 +549,23 @@ class Environment(RPC):
         negotiate_facades(self, self._info)
         return result
 
+    def _make_headers(self):
+        return {
+            'Authorization': 'Basic %s' % b64encode(
+                (u'%(user)s:%(password)s' % (self._creds)).encode()).decode()
+        }
+
     def _http_conn(self):
         endpoint = self.endpoint.replace('wss://', '')
         host, remainder = endpoint.split(':', 1)
-        port, _ = remainder.split('/', 1)
-        conn = HTTPSConnection(host, port)
-        headers = {
-            'Authorization': 'Basic %s' % b64encode(
-                '%(user)s:%(password)s' % (self._creds))}
+        port = remainder
+        if '/' in remainder:
+            port, _ = remainder.split('/', 1)
+        conn = HTTPSConnection(host, int(port))
         path = ""
         if self._env_uuid:
             path = "/environment/%s" % self._env_uuid
-        return conn, headers, path
+        return conn, self._make_headers(), path
 
     # Charm ops / see charm facade for listing charms in environ.
     def add_local_charm_dir(self, charm_dir, series):
@@ -516,7 +595,7 @@ class Environment(RPC):
             headers['Content-Length'] = size
         conn.request("POST", path, charm_file, headers)
         response = conn.getresponse()
-        result = json.loads(response.read())
+        result = json.loads(response.read().decode())
         if not response.status == 200:
             raise EnvError(result)
         return result
@@ -771,10 +850,6 @@ class Environment(RPC):
         sslopt = Connector.get_ssl_config(cert_pem)
 
         p = urllib.urlencode(d)
-        headers = [
-            'Authorization: Basic %s' % b64encode(
-                '%(user)s:%(password)s' % (self._creds))]
-
         if self._env_uuid:
             url = self.endpoint.rsplit('/', 1)[0]
             url += "/log"
@@ -785,7 +860,8 @@ class Environment(RPC):
             url = "%s?%s" % (url, p)
 
         conn = websocket.create_connection(
-            url, origin=self.endpoint, sslopt=sslopt, header=headers)
+            url, origin=self.endpoint, sslopt=sslopt,
+            header=self._make_headers())
 
         # Error message if any is pre-pended.
         result = json.loads(conn.recv())
@@ -803,14 +879,40 @@ class Environment(RPC):
             "Params": {"Commands": command,
                        "Timeout": timeout}})
 
-    def run(self, targets, command, timeout=None):
+    def run(self, command, timeout=None, machines=None,
+            services=None, units=None):
         """Run a shell command on the targets (services, units, or machines).
+        At least one target must be specified machines || services || units
         """
-        return self._rpc({
+
+        assert not (not machines and not services and not units), \
+            "You must specify a target"
+
+        rpc_dict = {
             "Type": "Client",
             "Request": "Run",
-            "Params": {"Commands": command,
-                       "Timeout": timeout}})
+            "Params": {
+                "Commands": command,
+                "Timeout": timeout,
+            }
+        }
+
+        if machines:
+            if not isinstance(machines, (list, tuple)):
+                machines = [machines]
+            rpc_dict["Params"].update({'Machines': machines})
+
+        if services:
+            if not isinstance(services, (list, tuple)):
+                services = [services]
+            rpc_dict["Params"].update({'Services': services})
+
+        if units:
+            if not isinstance(units, (list, tuple)):
+                units = [units]
+            rpc_dict["Params"].update({'Units': units})
+
+        return self._rpc(rpc_dict)
 
     # Machine ops
     def add_machine(self, series="", constraints=None,
@@ -972,6 +1074,8 @@ class Environment(RPC):
         if connection is None:
             watch_env = Environment(self.endpoint)
             watch_env.login(**self._creds)
+            if self._debug:
+                watch_env._debug = True
         else:
             watch_env = connection
 
@@ -1352,14 +1456,14 @@ class APIFacade(object):
     def _format_user_tag(self, n):
         if not n.startswith('user-'):
             n = "user-%s" % n
-        if not '@' in n:
+        if '@' not in n:
             n = "%s@local" % n
         return n
 
     def rpc(self, op):
-        if not 'Type' in op:
+        if 'Type' not in op:
             op['Type'] = self.name
-        if not 'Version' in op:
+        if 'Version' not in op:
             op['Version'] = self.version
         return self.env._rpc(op)
 
@@ -1439,7 +1543,7 @@ class UserManager(APIFacade):
         users = []
 
         for e in entities:
-            if not 'username' in e or not 'password' in e:
+            if 'username' not in e or 'password' not in e:
                 raise ValueError(
                     "Invalid parameter for set password %s" % entities)
             users.append(
@@ -1577,7 +1681,7 @@ class KeyManager(APIFacade):
 
     def add(self, user, keys):
         return self.rpc({
-            "Request": "ListKeys",
+            "Request": "AddKeys",
             "Params": {"User": user,
                        "Keys": keys}})
 
@@ -1765,7 +1869,7 @@ class HA(APIFacade):
         return self.rpc({
             'Request': 'EnsureAvailability',
             'Params': {
-                'EnvironTag': "environment-%s" % self._env_uuid,
+                'EnvironTag': "environment-%s" % self.env._env_uuid,
                 'NumStateServers': int(num_state_servers),
                 'Series': series,
                 'Constraints': self._format_constraints(constraints),
@@ -1802,7 +1906,7 @@ class Actions(APIFacade):
                 'Tag': '',
                 'Name': action_name,
                 'Receiver': u,
-                'Params': params})
+                'Parameters': params})
         return self._enqueue(actions)
 
     def _enqueue(self, actions):
@@ -2045,7 +2149,7 @@ class StatusTranslator(object):
         d.pop("Series")
         d.pop("CharmURL")
         name = d.pop('Name')
-        ports = d.pop('Ports')
+        ports = d.pop('Ports') or []
         tports = d.setdefault('Ports', [])
         # Workaround for lp:1425435
         if ports:
@@ -2070,7 +2174,7 @@ class StatusTranslator(object):
         for ep in d['Endpoints']:
             svc_rels = self.data.setdefault(
                 'services', {}).setdefault(
-                    ep['ServiceName'], {}).setdefault('relations', {})
+                ep['ServiceName'], {}).setdefault('relations', {})
             svc_rels.setdefault(
                 ep['Relation']['Name'], []).append(ep['RemoteService'])
 
